'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.List = undefined;

var _typeof2 = require('babel-runtime/helpers/typeof');

var _typeof3 = _interopRequireDefault(_typeof2);

var _regenerator = require('babel-runtime/regenerator');

var _regenerator2 = _interopRequireDefault(_regenerator);

var _iterator = require('babel-runtime/core-js/symbol/iterator');

var _iterator2 = _interopRequireDefault(_iterator);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

var _ord = require('./ord');

var _lib = require('./lib');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * A data constructor for a `List`.
 * @alias module:list.List
 * @kind class
 * @extends Type
 * @private
 */
/**
 * lazy-linked-lists
 * Lazy and infinite linked lists for JavaScript.
 *
 * source/list.js
 *
 * List data type.
 * @license ISC
 */

var List = exports.List = function () {
  /**
   * Create a new `List`.
   * @param {*} head - The value to put at the head of the `List`
   * @param {List} tail - The tail of the `List`, which is also a `List` (possibly the empty list)
   * @private
   */
  function List(x, xs) {
    (0, _classCallCheck3.default)(this, List);

    this.head = null;
    this.tail = null;
    this.head = function () {
      return x;
    };
    this.tail = function () {
      return xs;
    };
  }

  (0, _createClass3.default)(List, [{
    key: _iterator2.default,
    value: function value() {
      var listIterator = _regenerator2.default.mark(function listIterator(xs) {
        return _regenerator2.default.wrap(function listIterator$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return (0, _lib.head)(xs);

              case 2:
                xs = (0, _lib.tail)(xs);

              case 3:
                if (xs !== _lib.emptyList) {
                  _context.next = 0;
                  break;
                }

              case 4:
              case 'end':
                return _context.stop();
            }
          }
        }, listIterator, this);
      });
      var gen = listIterator(this);
      return {
        next: function next() {
          return gen.next();
        }
      };
    }
    /**
     * Determine whether the `List` on which it is called is exactly equal to another `List`.
     * @param {List} ys - The other `List` on which to perform the equality check
     * @returns {boolean} `true` if this `List` === `ys` and `false` otherwise
     */

  }, {
    key: 'isEq',
    value: function isEq(ys) {
      return (0, _lib.fromListToArray)(this).every(function (x, i) {
        return x === (0, _lib.fromListToArray)(ys)[i];
      });
    }
    /**
     * Compare the `List` on which it is called to another `List` to determine its relative "ordering"
     * (less than, greater than, or equal).
     * @param {List} ys - The other `List` to compare
     * @returns {Symbol} `LT`, `GT`, or `EQ`, the ordering of this `List` as compared to `ys`
     */

  }, {
    key: 'compare',
    value: function compare(ys) {
      if ((0, _lib.isEmpty)(this) && (0, _lib.isEmpty)(ys)) {
        return _ord.EQ;
      }
      if ((0, _lib.isEmpty)(this) && !(0, _lib.isEmpty)(ys)) {
        return _ord.LT;
      }
      if (!(0, _lib.isEmpty)(this) && (0, _lib.isEmpty)(ys)) {
        return _ord.GT;
      }
      if (this.head() === (0, _lib.head)(ys)) {
        return this.tail().compare((0, _lib.tail)(ys));
      }
      return this.head() < (0, _lib.head)(ys) ? _ord.LT : _ord.GT;
    }
    /**
     * Determine whether the `List` on which it is called is less than another `List`.
     * @param {List} ys - The other `List` to compare
     * @returns {boolean} `true` if this `List` < `ys` and `false` otherwise
     */

  }, {
    key: 'isLessThan',
    value: function isLessThan(ys) {
      return this.compare(ys) === _ord.LT;
    }
    /**
     * Determine whether the `List` on which it is called is less than or equal to another `List`.
     * @param {List} ys - The other `List` to compare
     * @returns {boolean} `true` if this `List` <= `ys` and `false` otherwise
     */

  }, {
    key: 'isLessThanOrEqual',
    value: function isLessThanOrEqual(ys) {
      return this.compare(ys) !== _ord.GT;
    }
    /**
     * Determine whether the `List` on which it is called is greater than another `List`.
     * @param {List} ys - The other `List` to compare
     * @returns {boolean} `true` if this `List` > `ys` and `false` otherwise
     */

  }, {
    key: 'isGreaterThan',
    value: function isGreaterThan(ys) {
      return this.compare(ys) === _ord.GT;
    }
    /**
     * Determine whether the `List` on which it is called is greater than or equal to another `List`.
     * @param {List} ys - The other `List` to compare
     * @returns {boolean} `true` if this `List` >= `ys` and `false` otherwise
     */

  }, {
    key: 'isGreaterThanOrEqual',
    value: function isGreaterThanOrEqual(ys) {
      return this.compare(ys) !== _ord.LT;
    }
    /**
     * Return the monoidal identity of a `List`, which is the empty list.
     * @returns {List} The empty list
     */

  }, {
    key: 'mempty',
    value: function mempty() {
      return _lib.emptyList;
    }
    /**
     * Perform an associative operation on the `List` on which it is called and another `List`. This
     * is a monoidal operation, which for this data type is equivalent to `listAppend`.
     * @param {List} ys - The other `List`
     * @returns {List} A new `List`, the result of the associative operation
     */

  }, {
    key: 'mappend',
    value: function mappend(ys) {
      return (0, _lib.listAppend)(this, ys);
    }
    /**
     * Apply a function to each value of the `List` on which it is called and accumulate the result.
     * @param {Function} f - The function to fold over the `List`
     * @param {*} acc - The starting accumulator value
     * @returns {*} The result of folding `f` over the `List`
     */

  }, {
    key: 'foldr',
    value: function foldr(f, acc) {
      if ((0, _lib.isEmpty)(this)) {
        return acc;
      }
      var x = this.head();
      var xs = this.tail();
      return f(x, xs.foldr(f, acc));
    }
    /**
     * Map each element of a `List` to a function, evaluate the functions from left to right, and
     * collect the results.
     * @param {Function} f - The function to map over the `List`
     * @returns {*} The result of the traversal
     */

  }, {
    key: 'traverse',
    value: function traverse(f) {
      var consf = function consf(x, ys) {
        return f(x).fmap(function (a) {
          return _lib.cons.bind(null, a);
        }).ap(ys);
      };
      return (0, _lib.isEmpty)(this) ? this.pure(_lib.emptyList) : this.foldr(consf, this.pure(_lib.emptyList));
    }
    /**
     * Map a function over each element of a `List`, and return the results in a new `List`.
     * @param {Function} f - The function to map over the `List`
     * @returns {List} A new `List`, the result of the mapping
     */

  }, {
    key: 'fmap',
    value: function fmap(f) {
      return (0, _lib.map)(f, this);
    }
    /**
     * Lift a value into the context of a `List`.
     * @param {*} a - The value to lift into a `List`
     * @returns {List} The lifted value
     */

  }, {
    key: 'pure',
    value: function pure(a) {
      return (0, _lib.list)(a);
    }
    /**
     * Apply the `List` of functions on which it is called to a `List` of values, and return the
     * results in a new `List`.
     * @param {List} ys - The `List` of values
     * @returns {List} A new `List`, the result of the applications
     */

  }, {
    key: 'ap',
    value: function ap(ys) {
      return (0, _lib.isEmpty)(this) ? _lib.emptyList : (0, _lib.listAppend)(ys.fmap(this.head()), this.tail().ap(ys));
    }
    /**
     * Perform a monadic "bind" operation. Apply a function that returns a `List` to the values of the
     * `List` on which it is called and flatten the result to produce a new `List`.
     * @param {Function} f - The function to apply to the values of this `List`
     * @returns {List} A new `List`, the result of the monadic operation
     */

  }, {
    key: 'flatMap',
    value: function flatMap(f) {
      return (0, _lib.concat)((0, _lib.map)(f, this));
    }
    /**
     * Perform a monadic chaining operation from one `List` to another, discarding the value of the
     * `List` on which it is called.
     * @param {List} ys - The function to chain
     * @returns {List} A new `List`, the result of the chaining operation
     */

  }, {
    key: 'then',
    value: function then(ys) {
      var c = function c(a, b) {
        return b, a;
      }; // eslint-disable-line
      var id = function id(a) {
        return a;
      };
      return this.fmap(c.bind(null, id)).ap(ys);
    }
  }, {
    key: 'toString',
    value: function toString() {
      return '[Object List]';
    }
  }, {
    key: 'typeOf',
    value: function typeOf() {
      return '[' + ((0, _lib.isEmpty)(this) ? '' : (0, _typeof3.default)(this.head())) + ']';
    }
  }, {
    key: 'valueOf',
    value: function valueOf() {
      if (this === _lib.emptyList) {
        return '[]';
      }
      var value = function value(xs) {
        return (0, _lib.isEmpty)(xs) ? '[]' : (0, _lib.head)(xs) + ':' + value((0, _lib.tail)(xs));
      };
      return '[' + (this.typeOf() === '[string]' ? (0, _lib.fromListToString)(this) : value(this)) + ']';
    }
  }]);
  return List;
}();